from __future__ import annotations
from typing import TYPE_CHECKING, Dict, List, Literal, Callable, Set, Tuple
from statistics import mean
from tibert.utils import spans_indexs

if TYPE_CHECKING:
    from tibert.bertcoref import CoreferenceDocument


def score_coref_predictions(
    preds: List[CoreferenceDocument], refs: List[CoreferenceDocument]
) -> Dict[
    Literal["MUC", "B3", "CEAF"],
    Dict[Literal["precision", "recall", "f1"], float],
]:
    """Score coreference prediction according to MUC, B3 and CEAF
    metrics

    .. note::

        Needs package ``neleval``

    :param preds: Predictions
    :param refs: References
    """
    assert len(preds) == len(refs)
    assert len(preds) > 0

    # neleval use np.int and np.bool, which are deprecated
    # (https://numpy.org/devdocs/release/1.20.0-notes.html#deprecations). The
    # two following lines fix the resulting crash.
    np.int = int  # type: ignore
    np.bool = bool  # type: ignore

    from neleval.coref_metrics import muc, b_cubed, ceaf

    def coref_doc_to_neleval_format(doc: CoreferenceDocument, max_span_size: int):
        """Convert a coreference document to the format expected by ``neleval``"""

        spans_idxs = spans_indexs(doc.tokens, max_span_size)

        # { chain_id => {tokens_id} }
        clusters = {}

        for chain_i, chain in enumerate(doc.coref_chains):

            clusters[chain_i] = set(
                [
                    spans_idxs.index((mention.start_idx, mention.end_idx))
                    for mention in chain
                ]
            )

        return clusters

    def precisions_recalls_f1s(
        preds: List[CoreferenceDocument],
        refs: List[CoreferenceDocument],
        neleval_fn: Callable[
            [Dict[int, Set[str]], Dict[int, Set[str]]],
            Tuple[float, float, float, float],
        ],
    ) -> Tuple[List[float], List[float], List[float]]:
        """Get precisions, recalls and f1s from a neleval metrics"""
        precisions = []
        recalls = []
        f1s = []
        for pred, ref in zip(preds, refs):
            try:
                pred_max_span_size = max(
                    [
                        (mention.end_idx - mention.start_idx) + 1
                        for chain in pred.coref_chains
                        for mention in chain
                    ]
                )
            except ValueError:
                pred_max_span_size = 0
            try:
                ref_max_span_size = max(
                    [
                        (mention.end_idx - mention.start_idx) + 1
                        for chain in ref.coref_chains
                        for mention in chain
                    ]
                )
            except ValueError:
                ref_max_span_size = 0
            max_span_size = max(pred_max_span_size, ref_max_span_size)
            # TODO max_span_size
            neleval_pred = coref_doc_to_neleval_format(pred, max_span_size + 1)
            neleval_ref = coref_doc_to_neleval_format(ref, max_span_size + 1)

            if neleval_pred == neleval_ref:
                precision = 1.0
                recall = 1.0
                f1 = 1.0
            else:
                # num = numerator
                # den = denominator
                p_num, p_den, r_num, r_den = neleval_fn(neleval_ref, neleval_pred)
                precision = p_num / p_den if p_den > 0 else 0.0
                recall = r_num / r_den if r_den > 0 else 0.0
                if precision + recall != 0.0:
                    f1 = 2 * precision * recall / (precision + recall)
                else:
                    f1 = 0.0

            precisions.append(precision)
            recalls.append(recall)
            f1s.append(f1)

        return precisions, recalls, f1s

    muc_precisions, muc_recalls, muc_f1s = precisions_recalls_f1s(preds, refs, muc)
    b3_precisions, b3_recalls, b3_f1s = precisions_recalls_f1s(preds, refs, b_cubed)
    ceaf_precisions, ceaf_recalls, ceaf_f1s = precisions_recalls_f1s(preds, refs, ceaf)

    return {
        "MUC": {
            "precision": mean(muc_precisions),
            "recall": mean(muc_recalls),
            "f1": mean(muc_f1s),
        },
        "B3": {
            "precision": mean(b3_precisions),
            "recall": mean(b3_recalls),
            "f1": mean(b3_f1s),
        },
        "CEAF": {
            "precision": mean(ceaf_precisions),
            "recall": mean(ceaf_recalls),
            "f1": mean(ceaf_f1s),
        },
    }



from typing import TYPE_CHECKING, Optional
import traceback
from statistics import mean
from more_itertools.recipes import flatten
import torch
from torch.utils.data.dataloader import DataLoader
from transformers import BertTokenizerFast  # type: ignore
from tqdm import tqdm
from tibert import (
    BertForCoreferenceResolution,
    BertCoreferenceResolutionOutput,
    CoreferenceDataset,
    split_coreference_document,
    DataCollatorForSpanClassification,
    score_coref_predictions,
)
from tibert.utils import gpu_memory_usage


def train_coref_model(
    model: BertForCoreferenceResolution,
    dataset: CoreferenceDataset,
    tokenizer: BertTokenizerFast,
    batch_size: int = 1,
    epochs_nb: int = 30,
    sents_per_documents_train: int = 11,
    bert_lr: float = 1e-5,
    task_lr: float = 2e-4,
    _run: Optional["sacred.run.Run"] = None,
) -> BertForCoreferenceResolution:

    if _run:
        from sacred.commands import print_config

        print_config(_run)

    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    train_dataset = CoreferenceDataset(
        dataset.documents[: int(0.9 * len(dataset))],
        dataset.tokenizer,
        dataset.max_span_size,
    )
    train_dataset.documents = list(
        flatten(
            [
                split_coreference_document(doc, sents_per_documents_train)
                for doc in train_dataset.documents
            ]
        )
    )

    test_dataset = CoreferenceDataset(
        dataset.documents[int(0.9 * len(dataset)) :],
        dataset.tokenizer,
        dataset.max_span_size,
    )
    test_dataset.documents = list(
        flatten(
            [
                # HACK: test on full documents
                split_coreference_document(doc, 200)
                for doc in test_dataset.documents
            ]
        )
    )

    data_collator = DataCollatorForSpanClassification(tokenizer, model.max_span_size)
    train_dataloader = DataLoader(
        train_dataset, batch_size=batch_size, shuffle=True, collate_fn=data_collator
    )
    test_dataloader = DataLoader(
        test_dataset, batch_size=batch_size, shuffle=False, collate_fn=data_collator
    )

    optimizer = torch.optim.AdamW(
        [
            {"params": model.bert_parameters(), "lr": bert_lr},
            {
                "params": model.task_parameters(),
                "lr": task_lr,
            },
        ],
        lr=task_lr,
    )

    best_f1 = 0
    best_model = model

    model = model.to(device)

    for epoch_i in range(epochs_nb):

        model = model.train()

        epoch_losses = []

        data_tqdm = tqdm(train_dataloader)
        for batch in data_tqdm:

            batch = batch.to(device)

            optimizer.zero_grad()

            try:
                out = model(**batch)
            except Exception as e:
                print(e)
                traceback.print_exc()
                continue

            assert not out.loss is None
            out.loss.backward()
            optimizer.step()

            _ = _run and _run.log_scalar("gpu_usage", gpu_memory_usage())

            data_tqdm.set_description(f"loss : {out.loss.item()}")
            epoch_losses.append(out.loss.item())
            if _run:
                _run.log_scalar("loss", out.loss.item())

        if _run:
            _run.log_scalar("epoch_mean_loss", mean(epoch_losses))

        # metrics computation
        model = model.eval()

        with torch.no_grad():

            try:

                preds = []
                losses = []

                for batch in tqdm(test_dataloader):

                    local_batch_size = batch["input_ids"].shape[0]
                    batch = batch.to(device)
                    out: BertCoreferenceResolutionOutput = model(**batch)
                    batch_preds = out.coreference_documents(
                        [
                            [tokenizer.decode(t) for t in batch["input_ids"][i]]
                            for i in range(local_batch_size)
                        ]
                    )
                    preds += batch_preds

                    assert not out.loss is None
                    losses.append(out.loss.item())

                _ = _run and _run.log_scalar("epoch_mean_test_loss", mean(losses))

                refs = [
                    doc.prepared_document(test_dataset.tokenizer, model.max_span_size)[
                        0
                    ]
                    for doc in test_dataset.documents
                ]
                metrics = score_coref_predictions(preds, refs)

                if _run:
                    _run.log_scalar("muc_precision", metrics["MUC"]["precision"])
                    _run.log_scalar("muc_recall", metrics["MUC"]["recall"])
                    _run.log_scalar("muc_f1", metrics["MUC"]["f1"])
                    _run.log_scalar("b3_precision", metrics["B3"]["precision"])
                    _run.log_scalar("b3_recall", metrics["B3"]["recall"])
                    _run.log_scalar("b3_f1", metrics["B3"]["f1"])
                    _run.log_scalar("ceaf_precision", metrics["CEAF"]["precision"])
                    _run.log_scalar("ceaf_recall", metrics["CEAF"]["recall"])
                    _run.log_scalar("ceaf_f1", metrics["CEAF"]["f1"])

                print(metrics)

                # keep the best model
                model_f1 = mean(
                    [metrics["MUC"]["f1"], metrics["B3"]["f1"], metrics["CEAF"]["f1"]]
                )

            except Exception as e:
                print(e)
                traceback.print_exc()
                model_f1 = 0

            if model_f1 > best_f1 or best_f1 == 0:
                best_model = model.to("cpu")
                best_f1 = model_f1

    return best_model
